
// project: webCDwriter - Network CD Writing
// file: Server.cpp
// author: Jrg Haeger, 21.04.1999

#include <arpa/inet.h>
#include <assert.h>
#include <crypt.h>
#include <dirent.h>
#include <errno.h>
#include <netdb.h>
#include <netinet/in.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/vfs.h>
#include <sys/wait.h>
#include <unistd.h>
#include <utime.h>

#include "Hints.h"
#include "Support.h"
#include "Process.h"
#include "Version.h"
#include "Log.h"
#include "Config.h"
#include "Queue.h"
#include "Session.h"
#include "Server.h"
#include "Store.h"
#include "Status.h"

const char
	hint[] = "HINT",
	info[] = "INFO",
	*unknown = "unknown",
	waitForDiscMsg1[] =
		"Der CD-Brenner wurde fr %d Minuten reserviert.",
	waitForDiscMsg2[] =
		"Bitte jetzt einen CD-R-Rohling einlegen ",
	waitForDiscMsg3[] =
		"und abwarten, bis er wieder ausgeworfen wird!",
	infoFileName[] = "/etc/CDWserver/greeting",
	waitForDiscMsgFileName[] = "/etc/CDWserver/waitForCD";

void readMultiSessionInfo(int &a, int &b);
char *removeEscChars(char *str);
const char *secondsToString(time_t t, char *str);
const char *timeToDigits(time_t t, char *str);
const char *timeToString(time_t t, char *str);

Server::Server(int socket)
{
	time(&connectionT0);
	char timeStr[80];
	log.put("%s started", timeToString(connectionT0, timeStr));

	activeServersNum++;
	serverNo = nextServerNo++;
	sock = socket;
	inStream = fdopen(sock, "r");
	logFilesNum = 0;

	// try to determine the requesting host
	struct sockaddr_in peerName;
	socklen_t peerNameLen = sizeof peerName;
	if (getpeername(sock, (struct sockaddr *)&peerName,
		&peerNameLen) == 0)
		if (peerName.sin_family == AF_INET)
			log.put("peerName <%s> port = %d\n",
				inet_ntoa(peerName.sin_addr),
				peerName.sin_port);

	// set the socket to SO_KEEPALIVE
	int on = 1;
	if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE,
		(void *)&on, sizeof on) < 0)
		log.put("setsockopt SO_KEEPALIVE: %s", strerror(errno));

	// states
	bytesReceived = 0;
	openCommands = 0;
	isOffline = 0;
	clientSystemName = clientSystemVersion = unknown;
	clientUserID = clientUserName = unknown;
	userID = userPassword = userID2 = unknown;

	// modes
	modeBurnOffline = 0;
	modeLaserOff = 0;
	modeShowCdrecord = 0;
	modeShowMkisofs = 0;

	session = NULL;
	quit = 0;

	toClient(info, "%s %s\nusing \"%s\"",
		product, version, config.getWriterInfo());

	FILE *inStream = fopen(infoFileName, "r");
	if (inStream != NULL)
	{
		while (1)
		{
			char str[1024];
			fgets(str, sizeof str, inStream);
			if (feof(inStream))
				break;
			toClient(info, str);
		}
		fclose(inStream);
	}

	openCommands++;
	if (configError)
	{
		char str[4 * 1024];
		gethostname(str, sizeof str);
		errorToClient("server configuration error"
			" - see http://%s:%d/ for details",
			str, config.getStatusPortNo());
		quit = 1;
	}
	else
		successToClient("connected");
}

int Server::activeServersNum = 0;
int Server::nextServerNo = 1;
Queue *Server::writerQueue = NULL;
int Server::writerStatus = 0;
int Server::writerPercent = 0;
int Server::configError = 0;

Server::~Server()
{
	close(sock);

	time_t t;
	time(&t);
	char timeStr[80];
	log.put("%s terminated", timeToString(t, timeStr));
	logConnection();

	if (session != NULL)
	{
		delete session;
		session = NULL;
	}

	activeServersNum--;
	writeStatusFile();
}

void Server::burnSession(char *cmdStr)
{
	// end of transfer
	time(&sessionT1);
	int hasWriter = 0;
	try {
		writerQueue->add();

		int cancel = waitForService(writerQueue, 0);
		if (cancel)
		{
			log.put("wait for reservation cancelled");
			throw "Warten auf Reservierung abgebrochen";
		}

		hasWriter = 1;
		strcat(cmdStr, " ");
		int multiTOC = strstr(cmdStr, " multi ") != NULL;
		modeBurnOffline = strstr(cmdStr, " offline ") != NULL;

		if (session->getFirstTrack() == NULL)
			throw "die Session enthlt keine Dateien";

		waitForDisc();

		if (session->getFirstTrack()->isDirectory())
		{
			toClient(hint, hintMakingImage);
			try {
				run_mkisofs();
				toClient(hint, hintImageReady);
				toClient(hint, hintCDrecordStarted);
				run_cdrecord(multiTOC);
			}
			catch (const char *msg) {
				logSession("S", msg);
				throw msg;
			}
			logSession("S", "OK");
		}
		else
		{
			toClient(hint, hintCDrecordStarted);
			try {
				run_cdrecord(multiTOC);
			}
			catch (const char *msg) {
				logSession("I", msg);
				throw msg;
			}
			logSession("I", "OK");
		}
		successToClient("Brennvorgang erfolgreich abgeschlossen");
		log.put("cdrecord erfolgreich");

		writerStatus = 0;
		writerQueue->remove();
	}
	catch (const char *msg) {
		if (hasWriter)
			writerStatus = 0;
		writerQueue->remove();
		throw msg;
	}
}

void Server::eject()
{
	Process p(config.getCdrecordPath());
	char str[4*1024];
	sprintf(str, "dev=%s", config.getCdrecordDevArg());
	p.addArg(str);
	p.addArg("-eject");

	p.start();
	while (1)
	{
		const char *lineStr = p.readLine();
		if (lineStr == NULL)
			break;
	}
	if (p.getResult() != 0)
		throw "can't eject";
}

void Server::errorToClient(const char *format, ...)
{
	if (isOffline)
		return;
	if (openCommands == 0)
	{
		log.put("error: openCommands = %d", openCommands);
		return;
	}

	char str[8*1024];
	va_list ap;
	va_start(ap, format);
	vsnprintf(str, sizeof str - 1, format, ap);
	va_end(ap);

	// Nachricht an Client besteht aus genau einer Zeile
	int i = 0;
	while (str[i] != 0 && str[i] != '\n')
		i++;
	str[i++] = '\n';
	str[i] = 0;

	char lineStr[16*1024];
	sprintf(lineStr, "ERR %s", str);
	int n = ::write(sock, lineStr, strlen(lineStr));
	openCommands--;
}

void Server::executeCmd()
{
	openCommands++;

	char str1[sizeof lineStr],
		str2[sizeof lineStr], str3[sizeof lineStr];
	int strsNum = sscanf(lineStr, "%s %s %s", str1, str2, str3);

	if (strsNum > 0
		&& (strcasecmp(str1, "putFile") == 0
			|| strcasecmp(str1, "mkdir") == 0))
		log.putv("command: <%s>", lineStr);
	else
		log.put("command: <%s>", lineStr);

	if (strlen(lineStr) == 0)
		errorToClient("empty command");
	else
	if (strcasecmp(str1, "client") == 0)
	{
		successToClient("");
	}
	else
	if (strcasecmp(str1, "login") == 0)
	{
		if (strsNum == 3)
		{
			if (userID == unknown)
				userID = strDup(str2);
			userID2 = strDup(str2);
			userPassword = strDup(crypt(str3, "JH"));
			successToClient("LOGIN completed");
		}
		else
			errorToClient("in command %s", str1);
	}
	else
	if (strcasecmp(str1, "newSession") == 0
		|| strcasecmp(str1, "openSession") == 0)
	{
		if (strsNum <= 2)
		{
			int len = strlen(userID2)
				+ 1 + strlen(userPassword);
			if (strsNum == 2)
				len += 1 + strlen(str2);

			char sessionID[len + 1];
			sprintf(sessionID, "%s-%s",
				userID2, &userPassword[2]);
			if (strsNum == 2)
				strcat(sessionID, str2);

			try
			{
				if (session != NULL)
					delete session;
				session = new Session(this, sessionID);
				if (strcasecmp(str1, "openSession") == 0)
					session->open();
				if (session->getFilesNum() > 0)
					toClient(hint, "files %d",
						session->getFilesNum());
				else
				if (session->getTracksNum() > 0)
					toClient(hint, "files %d",
						session->getTracksNum());
				successToClient("");
				time(&sessionT0);
			}
			catch (const char *msg)
			{
				log.put(msg);
				errorToClient(msg);
			}
			writeStatusFile();
		}
		else
			errorToClient("in command %s", str1);
	}
	else
	if (strcasecmp(str1, "noop") == 0)
		successToClient("");
	else
	if (strcasecmp(str1, "quit") == 0)
	{
		log.put("quit von client");
		successToClient("");
		quit = 1;
	}
	else
	if (strcasecmp(str1, "setShowCdrecord") == 0)
	{
		modeShowCdrecord = 1;
		successToClient("");
	}
	else
	if (strcasecmp(str1, "setLaserOff") == 0)
	{
		modeLaserOff = 1;
		successToClient("");
	}
	else
	if (strcasecmp(str1, "setShowMkisofs") == 0)
	{
		modeShowMkisofs = 1;
		successToClient("");
	}
	else
	if (strcasecmp(str1, "system") == 0)
	{
		if (strsNum >= 2)
			clientSystemName = strDup(str2);
		if (strsNum >= 3)
			clientSystemVersion = strDup(str3);
		successToClient("");
	}
	else
	if (strcasecmp(str1, "user") == 0)
	{
		if (strsNum >= 2 && strcasecmp(str2, "noID") != 0)
			clientUserID = strDup(str2);
		if (strsNum >= 3)
			clientUserName = strDup(str3);
		if (userID == unknown)
		{
			char *str = new char[
				strlen(clientSystemName)
				+ 1 + strlen(clientUserID)
				+ 1 + strlen(clientUserName) + 1];
			if (clientUserID == unknown)
				sprintf(str, "%s-%s", clientSystemName,
					clientUserName);
			else
				sprintf(str, "%s-%s-%s", clientSystemName,
					clientUserID, clientUserName);
			userID = str;
		}
		successToClient("");
	}
	else
	if (strcasecmp(str1, "vm") == 0)
	{
		successToClient("");
	}
	else
	if (session == NULL)
	{
		errorToClient("keine Session offen");
	}
	else
	if (strcasecmp(str1, "burnSession") == 0)
		burnSession(lineStr);
	else
	if (strcasecmp(str1, "closeSession") == 0)
	{
		try
		{
			delete session;
			session = NULL;
			successToClient("");
		}
		catch (const char *msg)
		{
			log.put(msg);
			errorToClient(msg);
		}
	}
	else
	if (strcasecmp(str1, "mkdir") == 0)
	{
		char fileName[sizeof lineStr];
		int mode, time;
		sscanf(lineStr, "mkdir %s %o %d",
			fileName, &mode, &time);
		removeEscChars(fileName);

		session->makeDir(fileName, mode, time);
		successToClient("Verzeichnis angelegt");
	}
	else
/*	if (strcasecmp(str1, "mklink") == 0)
	{
		char fileName[sizeof lineStr];
		int mode, time;
		sscanf(lineStr, "mklink %s %o %d",
			fileName, &mode, &time);
		removeEscChars(fileName);

		try
		{
			session->mklink(fileName, mode, time);
			successToClient("Link angelegt");
		}
		catch (const char *msg)
		{
			errorToClient(msg);
		}
	}
	else
*/	if (strcasecmp(str1, "putFile") == 0)
	{
		char fileName[1024];
		int length, mode, time;
		if (sscanf(lineStr, "%s %s %d %o %d",
			str1, fileName, &length, &mode, &time) != 5)
			throw "Fehler im Befehl";
		removeEscChars(fileName);
		bytesReceived += length;
		session->putFile(inStream, fileName, length, mode, time);
		successToClient("Datei gespeichert");
	}
	else
	if (strcasecmp(str1, "putTrack") == 0)
	{
		char trackType[1024];
		int length;
		if (sscanf(lineStr, "%s %s %d",
			str1, trackType, &length) != 3)
			throw "Fehler im Befehl";
		bytesReceived += length;
		if (strcmp(trackType, "continue") != 0)
			session->newTrack(trackType);
		session->putTrack(inStream, length);
		successToClient("Datei gespeichert");
	}
	else
		errorToClient("unknown command");
}

const char *Server::getNewLogFileName(char *str)
{
	char timeStr[80];
	sprintf(str, "%s/%s/%s.%03d",
		config.getLogDir(), userID,
		timeToDigits(connectionT0, timeStr),
		++logFilesNum);
	makeDirs(str);
	return str;
}

void Server::initStatic()
{
	char *PATH = getenv("PATH");
	if (PATH == NULL)
		PATH = "";
	log.put("PATH=%s", getenv("PATH"));

	config.readConfigFile();
	configError = config.getFirstError() != NULL;
	status.start();
	if (!configError)
		initStatic2();
}

void Server::initStatic2()
{
	configError = config.getFirstError() != NULL;

	if (writerQueue == NULL)
		writerQueue = new Queue(config.getMaxOpenSessions());

	log.put("used CD-writer: \"%s\"", config.getWriterInfo());

	try {
		store = new Store();
	}
	catch (const char *msg)
	{
		log.put("%s", msg);
	}
}

void Server::load()
{
	Process p(config.getCdrecordPath());
	char str[4*1024];
	sprintf(str, "dev=%s", config.getCdrecordDevArg());
	p.addArg(str);
	p.addArg("-load");

	p.start();
	while (1)
	{
		const char *lineStr = p.readLine();
		if (lineStr == NULL)
			break;
	}
	if (p.getResult() == 0)
		eject();
}

void Server::logConnection()
{
	char fileName[4*1024];
	sprintf(fileName, "%s/connects", config.getLogDir());
	FILE *file = fopen(fileName, "a");
	assert(file != NULL);

	time_t t;
	time(&t);
	char timeStr[80];
	fprintf(file,
		"%s %-20s ",
		timeToString(connectionT0, timeStr), userID);
	fprintf(file, "%s %10d\n",
		secondsToString(t - connectionT0, timeStr),
		bytesReceived);
	fclose(file);
}

void Server::logSession(const char *typeStr, const char *resultStr)
{
	char fileName[4*1024];
	sprintf(fileName, "%s/sessions", config.getLogDir());
	FILE *file = fopen(fileName, "a");
	if (file == NULL)
		return;

	time_t t;
	time(&t);
	char timeStr[80];
	fprintf(file,
		"%s %-20s ",
		timeToString(connectionT0, timeStr), userID);
	fprintf(file, "%s ",
		secondsToString(sessionT1 - sessionT0, timeStr));
	fprintf(file, "%s ", secondsToString(t - burnT0, timeStr));
	fprintf(file, "%10d %s %s\n",
		session->getSize(), typeStr, resultStr);
	fclose(file);

	if (strcasecmp(resultStr, "OK") != 0
		&& config.getEMailOnError() != NULL)
	{
		char cmdStr[4*1024];
		sprintf(cmdStr, "echo %s %s %s %s | mail -s %s %s",
			timeToString(connectionT0, timeStr),
			userID, typeStr, resultStr,
			"CD-Writer-Error", config.getEMailOnError());
		system(cmdStr);
	}

	timeToString(t, timeStr);
	char date[20], time2[20];
	strncpy(date, timeStr, 8);
	date[8] = 0;
	strncpy(time2, &timeStr[9], 9);
	time2[9] = 0;
	status.addResult(date, time2, session->getSize(), resultStr);
	writeStatusFile();
}

int Server::nextCmd()
{
	if (fgets(lineStr, sizeof lineStr, inStream) == NULL)
		return 0;

	int l = 0;
	while (lineStr[l] != 0 && lineStr[l] != 10 && lineStr[l] != 13)
		l++;
	lineStr[l] = 0;

	l = 0;
	while (((unsigned char *)lineStr)[l] >= 32)
		l++;
	if (lineStr[l] == 0)
		try {
			executeCmd();
		}
		catch (const char *msg) {
			errorToClient(msg);
			log.put("error: %s", msg);
		}
	else
	{
		log.put("Kommando enthlt Steuerzeichen");
		errorToClient("Kommando enthlt Steuerzeichen");
		quit = 1;
	}

	return 1;
}

int Server::noDisc()
{
//	return recorder->noDisc();
	Process p(config.getCdrecordPath());

	// general options to cdrecord
	char str[4*1024];
	sprintf(str, "dev=%s", config.getCdrecordDevArg());
	p.addArg(str);
	p.addArg("-inq");
	p.addArg("-V");

	p.start();
	int errorCode = 0;
	while (1)
	{
		const char *lineStr = p.readLine();
		if (lineStr == NULL)
			break;

		if (strstr(lineStr, "medium not present") != NULL)
			errorCode = 1;
	}

	return p.getResult() != 0 || errorCode == 1;
}

void Server::run()
{
	while (!quit && nextCmd());
}

void Server::run_cdrecord(int multiTOC)
{
	writerPercent = 0;
	writerStatus = 3;
	time(&burnT0);
	log.putv("Server::run_cdrecord");
	char str[4*1024];
	Process p(config.getCdrecordPath(), getNewLogFileName(str));

	// general options to cdrecord
	sprintf(str, "dev=%s", config.getCdrecordDevArg());
	p.addArg(str);
	if (modeLaserOff || !config.getLaserOn())
		p.addArg("-dummy");
	p.addArg("-eject");
	CD.isMultiSession = multiTOC && CD.extentsWrittenKB < 628*1024;
	if (CD.isMultiSession)
		p.addArg("-multi");
	sprintf(str, "speed=%d", config.getSpeed());
	p.addArg(str);
	p.addArg("-v");

	// track options/tracks to cdrecord
	const File *track = session->getFirstTrack();
	if (track->isDirectory())
		p.addArg(config.getImageFile());
	else
		while (track != NULL)
		{
			p.addArg(track->getName());
			track = session->getNextTrack();
		}

	p.show();
//	log.putv(p.toString(str));

	p.start();
	int errorCode = 0;
	while (1)
	{
		const char *lineStr = p.readLine();
		if (lineStr == NULL)
			break;
		toClient("CDRECORD", "%s", lineStr);

		if (strstr(lineStr, "Blocks total:") == lineStr)
		{
			CD.scan(lineStr);
			if (CD.valid)
				toClient(hint, "capacity %d %d %d",
				CD.totalKB, CD.oldFreeKB, CD.newFreeKB);
		}
		else
		if (strncmp(lineStr, "Track", 5) == 0)
		{
			toClient(hint, lineStr);
			int trackNo, a, b;
			if (sscanf(lineStr, "Track %d: %d of %d",
				&trackNo, &a, &b) == 3)
				writerPercent = 100 * a / b;
		}
		else
		if (strstr(lineStr, "may not fit") != NULL)
			errorCode = 1;
		else
		if (strstr(lineStr, "Drive needs to reload") != NULL)
			// the media to return to proper status.
			errorCode = 2;
	}

	if (p.getResult() != 0)
	switch (errorCode)
	{
		case 1:
			throw "nicht genug Platz auf der CD";
		case 2:
			throw "Brennvorgang mir der selben CD wiederholen";
		default:
			throw "Fehler beim Brennen";
	}
}

void Server::run_mkisofs()
{
	writerPercent = 0;
	writerStatus = 2;
	log.putv("Server::run_mkisofs %s",
		session->getFirstTrack()->getName());
	char str[4*1024];
	Process p(config.getMkisofsPath(), getNewLogFileName(str));

	// arguments for mkisofs
	p.addArg("-all-files");
	p.addArg("-joliet");
	p.addArg("-full-iso9660-filenames");
	p.addArg("-allow-leading-dots");
	p.addArg("-rational-rock");
	p.addArg("-verbose");
	p.addArg("-gui");
	p.addArg("-o");
	p.addArg(config.getImageFile());
	p.addArg("-p");
	char timeStr[80];
	sprintf(str, "Org: %s, User: %s, %s, %s %s",
		config.getOrganization(), userID,
		timeToString(sessionT1, timeStr),
		product, version);
	p.addArg(str);

	try {
		int a, b;
		readMultiSessionInfo(a, b);
		p.addArg("-C");
		sprintf(str, "%d,%d", a, b);
		p.addArg(str);
		p.addArg("-M");
		p.addArg(config.getMkisofsMArg());
		toClient(hint, "msinfo %d %d", 2*a, 2*b);
	}
	catch (const char *msg) {
		FILE *rawCDStream = fopen(config.getMkisofsMArg(), "r");
		if (rawCDStream != NULL)
		{
			char buf[2*1024];
			int n = fread(buf, 1, sizeof buf, rawCDStream);
			fclose(rawCDStream);
			if (n == sizeof buf)
			{
				// can't get msinfo although CD isn't empty
				eject();
				throw "Brenner kann den Zustand"
					" der CD nicht bestimmen";
			}
		}
		log.putv(msg);
		toClient(hint, "msinfo 0 0");
	}

	p.addArg(".");

	p.show();
//	log.putv(p.toString(str));

	p.start(session->getFirstTrack()->getName());
	while (1)
	{
		const char *lineStr = p.readLine();
		if (lineStr == NULL)
			break;
		toClient("MKISOFS", "%s", lineStr);

		char *str;
		long int n = strtol(lineStr, &str, 10);
		if (str == lineStr)
			continue;

		if (strstr(str, " extents written") == str)
		{
			CD.extentsWrittenKB = 2 * n;
			log.put("mkisofs extentsWritten = %d MB",
				CD.extentsWrittenKB / 1024);
		}
		if (*str == ' ')
			toClient(hint, "KByte %ld", 2 * n);
		if (*str == '.')
		{
			float percent;
			if (sscanf(lineStr, "%f", &percent) == 1)
			{
				toClient(hint,
					"mkisofsPercentDone %f", percent);
				writerPercent = (int)percent;
			}
		}
	}

	if (p.getResult() != 0)
	{
		eject();
		throw "konnte Image nicht erzeugen";
	}
}

void Server::showWaitForDiscMsg()
{
	FILE *inStream = fopen(waitForDiscMsgFileName, "r");
	if (inStream == NULL)
	{
		toClient(info, waitForDiscMsg1,
			config.getWaitForCD());
		toClient(info, waitForDiscMsg2);
		toClient(info, waitForDiscMsg3);
	}
	else
	{
		while (1)
		{
			char str[1024];
			fgets(str, sizeof str, inStream);
			if (feof(inStream))
				break;
			toClient(info, str, config.getWaitForCD());
		}
		fclose(inStream);
	}
}

void Server::successToClient(const char *format, ...)
{
	if (isOffline)
		return;
	if (openCommands == 0)
	{
		log.put("error: openCommands = %d", openCommands);
		return;
	}

	char str[8*1024];
	va_list ap;
	va_start(ap, format);
	vsnprintf(str, sizeof str - 1, format, ap);
	va_end(ap);

	// Nachricht an Client besteht aus genau einer Zeile
	int i = 0;
	while (str[i] != 0 && str[i] != '\n')
		i++;
	str[i++] = '\n';
	str[i] = 0;

	char lineStr[16*1024];
	sprintf(lineStr, "OK %s", str);
	int n = ::write(sock, lineStr, strlen(lineStr));
	openCommands--;
}

void Server::terminate()
{
}

void Server::toClient(const char *type, const char *format, ...)
{
	if (isOffline
		|| (!modeShowCdrecord && strcmp(type, "CDRECORD") == 0)
		|| (!modeShowMkisofs && strcmp(type, "MKISOFS") == 0))
		return;

	char str[8*1024];
	va_list ap;
	va_start(ap, format);
	vsnprintf(str, sizeof str - 1, format, ap);
	va_end(ap);

	// Information an Client besteht aus genau einer Zeile
	int i = 0;
	while (str[i] != 0 && str[i] != '\n')
		i++;
	str[i++] = '\n';
	str[i] = 0;

	char lineStr[16*1024];
	sprintf(lineStr, "%s %s", type, str);
	int n = ::write(sock, lineStr, strlen(lineStr));
}

int Server::waitForDisc()
{
	eject();

	showWaitForDiscMsg();

	if (modeBurnOffline)
		successToClient("burnning offline");
	else
		toClient(hint, hintWriterReserved);
	isOffline = modeBurnOffline;

	log.put("writer reserved - waiting for CD");
	writerStatus = 1;

	struct timeval t0, t;
	gettimeofday(&t0, NULL);
	gettimeofday(&t, NULL);

	// wait for timeout, CD or new command on socket
	int cancel = 0;
	while (t.tv_sec-t0.tv_sec < config.getWaitForCD()*60
		&& noDisc() && !cancel)
	{
		if (modeBurnOffline)
			sleep(5);
		else
			cancel = waitForInput(inStream, 5) != 0;
		gettimeofday(&t, NULL);
	}

	if (cancel)
	{
		log.put("wait for CD cancelled");
		if (store->getSessionsNum() < 2)
			load();
		throw "Warten auf CD abgebrochen";
	}
	else
	if (noDisc())
	{
		log.put("reservation timeout");
		if (store->getSessionsNum() < 2)
			load();
		throw "Reservierung abgelaufen";
	}
	else
	{
		toClient(hint, hintCDdetected);
		return 0;
	}
}

int Server::waitForService(Queue *queue, int toRank)
{
	log.putv("Server::waitForService");

	// wait for rank 0 or cancel by new command on socket
	int rank, oldRank;
	rank = queue->getRank();
	oldRank = rank + 1;
	int cancel = 0;
	while (rank > toRank && !cancel)
	{
		if (rank < oldRank)
			toClient(hint, "rank %d", rank - toRank);
		cancel = waitForInput(inStream, 1) != 0;
		oldRank = rank;
		rank = queue->getRank();
	}
	return cancel;
}

void Server::writeStatusFile()
{
}


void readMultiSessionInfo(int &a, int &b)
{
	Process p(config.getCdrecordPath());
	char str[4*1024];
	sprintf(str, "dev=%s", config.getCdrecordDevArg());
	p.addArg(str);
	p.addArg("-msinfo");

	p.show();
//	log.putv(p.toString(str));

	p.start();
	const char *lineStr = p.readLine();
	if (sscanf(lineStr, "%d,%d", &a, &b) != 2)
		throw "blank disk";
	if (p.getResult() != 0)
		throw "kann Multisession-Information nicht lesen";
}

/**
 *	"\_" in " " und "\"ch in ch umwandeln
*/
char *removeEscChars(char *str)
{
	int i, j;
	for (i = 0, j = 0; str[i] != 0; i++, j++)
	{
		if (str[i] == '\\')
		{
			i++;
			if (str[i] == '_')
				str[i] = ' ';
		}
		str[j] = str[i];
	}
	str[j] = 0;
	return str;
}

const char *secondsToString(time_t t, char *str)
{
	struct tm *tm = gmtime(&t);
	sprintf(str, "%02d:%02d:%02d",
		tm->tm_hour, tm->tm_min, tm->tm_sec);
	return str;
}

const char *timeToDigits(time_t t, char *str)
{
	struct tm *tm = localtime(&t);
	sprintf(str, "%4d%02d%02d%02d%02d%02d",
		1900 + tm->tm_year, tm->tm_mon + 1, tm->tm_mday,
		tm->tm_hour, tm->tm_min, tm->tm_sec);
	return str;
}

const char *timeToString(time_t t, char *str)
{
	struct tm *tm = localtime(&t);
	sprintf(str, "%4d%02d%02d %02d:%02d:%02d",
		1900 + tm->tm_year, tm->tm_mon + 1, tm->tm_mday,
		tm->tm_hour, tm->tm_min, tm->tm_sec);
	return str;
}

